#pragma once

#include <csrmath.h>
#include <csrmath_rational.h>

namespace csrmath {

class ComplexD {
private:
	double re, im;

public:
	inline ComplexD() { };

	inline ComplexD(double re) : re(re), im(0.0) { };

	inline ComplexD(double re, double im) : re(re), im(im) { };

	inline double Re() const {
		return re;
	};

	inline double Im() const {
		return im;
	};

	inline ComplexD operator- () const {
		return ComplexD(-re,-im);
	}

	inline ComplexD operator~ () const {
		return ComplexD(re,-im);
	}

	inline ComplexD neg() const {
		return ComplexD(-re,-im);
	}

	inline ComplexD conj() const {
		return ComplexD(re,-im);
	}
	inline double abs() const {
		return csrmath::sqrt( re*re + im*im );
	}

	inline double abs2() const {
		return re*re + im*im;
	}

	inline double arg() const {
		return atan2( im, re );
	}

	inline ComplexD operator+ (ComplexD x) const {
		return ComplexD(re+x.re,im+x.im);
	}

	inline ComplexD operator- (ComplexD x) const {
		return ComplexD(re-x.re,im-x.im);
	}

	inline ComplexD operator* (ComplexD x) const {
		return ComplexD(
			re*x.re - im*x.im,
			re*x.im + im*x.re );
	}

	inline ComplexD operator/ (ComplexD x) const {
		double den = 1.0 / ( x.re*x.re + x.im*x.im );
		return ComplexD(
			( re*x.re + im*x.im ) * den,
			( im*x.re - re*x.im ) * den );
	}

	inline ComplexD operator+= (ComplexD x) {
		re += x.re;
		im += x.im;
		return *this;
	}

	inline ComplexD operator-= (ComplexD x) {
		re -= x.re;
		im -= x.im;
		return *this;
	}

	inline ComplexD operator*= (ComplexD x) {
		*this = (*this) * x;
		return *this;
	}

	inline ComplexD operator/= (ComplexD x) {
		*this = (*this) / x;
		return *this;
	}

	inline bool operator==(ComplexD x) const {
		return re==x.re && im==x.im;
	}

	inline bool operator!=(ComplexD x) const {
		return re!=x.re || im!=x.im;
	}

	inline ComplexD pow(int r) const;
	inline ComplexD pow(Rational r) const;
	inline ComplexD pow(Rational r, unsigned int n) const;
};

inline ComplexD operator* (double x, ComplexD y) {
	return y * x;
}

inline ComplexD expi(double arg) {
	return ComplexD( cos(arg), sin(arg) );
}

inline ComplexD exp(ComplexD x) {
	return expi( x.Im() ) * csrmath::exp( x.Re() );
}

inline ComplexD sqrt(ComplexD x) {
	return expi( x.arg()/2.0 ) * csrmath::sqrt( x.abs() );
}

inline ComplexD log(ComplexD x) {
	return ComplexD( csrmath::log( x.abs() ), x.arg() );
}

ComplexD ComplexD::pow(int n) const {
	if( n==0 )
		return 1.0;
	if( n>0 ) {
		ComplexD x = *this;
		for( n--; n>0; n-- )
			x *= *this;
		return x;
	} else {
		ComplexD x = ComplexD(1.0) / *this;
		for( n++; n<0; n++ )
			x /= *this;
		return x;
	}
}

ComplexD ComplexD::pow(Rational r) const {
	return pow(r,0);
}

ComplexD ComplexD::pow(Rational r, unsigned int n) const {
	return expi( arg()*(double)r + (2.0*Const::Pi)*n/r.den ) * csrmath::pow(abs(),r);
}

}
