#pragma once

#include <csrmath.h>

namespace csrmath {

class MatrixIncompatibleShape : public std::exception {
public:
	const char* what() {
		return "Matrices have incompatible shapes";
	}
};

class MatrixNonSquare : public std::exception {
public:
	const char* what() {
		return "Matrix isn't square";
	}
};

struct MatrixD_Impl;

void MatrixD_Impl_free(MatrixD_Impl*);

class MatrixD {
	size_t H, W;
	double *data;

public:
	MatrixD_Impl *impl;

	inline MatrixD() : W(0), H(0), data(nullptr), impl(nullptr) { }

	inline MatrixD(size_t _H, size_t _W) : W(_W), H(_H), impl(nullptr) {
		data = MATH_ALLOCS(double,W*H);
	}

	inline MatrixD(size_t _H, size_t _W, const double* _matrix) : W(_W), H(_H), impl(nullptr) {
		data = MATH_ALLOCS(double,W*H);
		for( size_t i=0; i<W*H; i++ )
			data[i] = _matrix[i];
	}

	inline MatrixD(const MatrixD& mat) : W(mat.W), H(mat.H), impl(nullptr) {
		data = MATH_ALLOCS(double,W*H);
		for( size_t i=0; i<W*H; i++ )
			data[i] = mat.data[i];
	}

	inline MatrixD(MatrixD&& mat) : W(mat.W), H(mat.H) {
		data = mat.data;
		impl = mat.impl;
		mat.data = nullptr;
		mat.impl = nullptr;
	}

	inline ~MatrixD() {
		if( data != nullptr ) {
			MATH_FREE(data);
			data = nullptr;
		}
		if( impl != nullptr ) {
			MatrixD_Impl_free(impl);
			impl = nullptr;
		}
	}

	inline size_t getRows() const { return H; }

	inline size_t getCols() const { return W; }

	inline MatrixD& operator= (const MatrixD& mat) {
		if( data!=nullptr && ( W!=mat.W || H!=mat.H ) ) {
			MATH_FREE(data);
			data = nullptr;
		}
		if( impl != nullptr ) {
			MatrixD_Impl_free(impl);
			impl = nullptr;
		}
		W = mat.W;
		H = mat.H;
		if( data == nullptr )
			data = MATH_ALLOCS(double,W*H);
		for( size_t i=0; i<W*H; i++ )
			data[i] = mat.data[i];
	}

	inline MatrixD& operator= (MatrixD&& mat) {
		if( data!=nullptr )
			MATH_FREE(data);
		if( impl != nullptr )
			MatrixD_Impl_free(impl);
		W = mat.W;
		H = mat.H;
		data = mat.data;
		impl = mat.impl;
		mat.data = nullptr;
		mat.impl = nullptr;
		return *this;
	}

	inline bool operator== (const MatrixD& mat) const {
		if( H!=mat.H || W!=mat.W )
			return false;
		const double epsilon = 0.001;
		for( size_t i=0; i<W*H; i++ )
			//if( data[i] != mat.data[i] )
			if( abs( data[i] - mat.data[i] ) > epsilon )
				return false;
		return true;
	}

	inline bool operator!= (const MatrixD& mat) const {
		return !(*this==mat);
	}

	inline double& get(size_t i, size_t j) {
		if( i>=H || j>=W )
			throw 0;
		return data[i*W+j];
	}

	inline double get(size_t i, size_t j) const {
		return data[i*W+j];
	}

	class Row {
		MatrixD& matrix;
		size_t row;
	public:
		inline Row(MatrixD& matrix, size_t row) : matrix(matrix), row(row) { }
		inline double& operator[] (size_t j) {
			return matrix.get(row,j);
		}
	};

	class ConstRow {
		const MatrixD& matrix;
		size_t row;
	public:
		inline ConstRow(const MatrixD& matrix, size_t row) : matrix(matrix), row(row) { }
		inline double operator[] (size_t j) const {
			return matrix.get(row,j);
		}
	};

	inline Row operator[] (size_t i) {
		return Row(*this,i);
	}

	inline ConstRow operator[] (size_t i) const {
		return ConstRow(*this,i);
	}

	inline MatrixD operator+ (const MatrixD& mat) const {
		if( W!=mat.W || H!=mat.H )
			throw MatrixIncompatibleShape();
		MatrixD res(W,H);
		for( size_t i=0; i<W*H; i++ )
			res.data[i] = data[i] + mat.data[i];
		return res;
	}

	inline MatrixD operator- (const MatrixD& mat) const {
		if( W!=mat.W || H!=mat.H )
			throw MatrixIncompatibleShape();
		MatrixD res(H,W);
		for( size_t i=0; i<W*H; i++ )
			res.data[i] = data[i] - mat.data[i];
		return res;
	}

	inline void operator+= (const MatrixD& mat) const {
		if( W!=mat.W || H!=mat.H )
			throw MatrixIncompatibleShape();
		for( size_t i=0; i<W*H; i++ )
			data[i] += mat.data[i];
	}

	inline void operator-= (const MatrixD& mat) const {
		if( W!=mat.W || H!=mat.H )
			throw MatrixIncompatibleShape();
		for( size_t i=0; i<W*H; i++ )
			data[i] -= mat.data[i];
	}

	inline MatrixD operator* (const MatrixD& mat) const {
		if( W!=mat.H )
			throw MatrixIncompatibleShape();
		MatrixD res(H,mat.W);
		for( size_t i=0; i<res.H; i++ )
		for( size_t j=0; j<res.W; j++ ) {
			double S = 0.0;
			for( size_t k=0; k<W; k++ )
				S += get(i,k) * mat.get(k,j);
			res.get(i,j) = S;
		}
		return res;
	}

	MatrixD inverse() const;
	MatrixD inplace_inverse();

	void decomposeLU(MatrixD& L, MatrixD& U);

	double det() const;

	void rowReduce();
};

#ifdef _IOSTREAM_
std::ostream& operator<< (std::ostream &sout, const MatrixD& mat) {
	size_t I = mat.getRows(), J = mat.getCols();
	for( size_t i=0; i<I; i++ )
	for( size_t j=0; j<J; j++ )
			sout << mat.get(i,j) << ( j==J-1 ? '\n' : ' ' );
	return sout;
}
#endif

}
