#pragma once

#include <csrmath.h>
#include <csrmath_numeric.h>

namespace csrmath {

//__declspec(thread) int tls_i = 1;

class Polynomial {
	double *coefs;
	size_t count;

public:
	inline Polynomial() : coefs(NULL), count(0) { }

	inline Polynomial(size_t count, double *_coefs)
			//: coefs( new double[count] ), count(count) {
			: coefs( MATH_ALLOCS(double,count) ), count(count) {
		memcpy(coefs,_coefs,sizeof(double)*count);
	}

	inline Polynomial(const Polynomial& p)
			//: coefs( new double[p.count] ), count(p.count) {
			: coefs( MATH_ALLOCS(double,p.count) ), count(p.count) {
		memcpy(coefs,p.coefs,sizeof(double)*p.count);
	}

	inline Polynomial(Polynomial&& p)
			: coefs(p.coefs), count(p.count) {
		p.coefs = NULL;
		p.count = 0;
	}

	inline ~Polynomial() {
		if( count > 0 )
			//delete[] coefs;
			MATH_FREE(coefs);
	}

	inline Polynomial& operator= (const Polynomial& p) {
		if( count > 0 )
			//delete[] coefs;
			MATH_FREE(coefs);
		count = p.count;
		//coefs = new double[p.count];
		coefs = MATH_ALLOCS(double,p.count);
		memcpy(coefs,p.coefs,sizeof(double)*p.count);
		return *this;
	}

	inline Polynomial& operator= (Polynomial&& p) {
		if( count > 0 )
			//delete[] coefs;
			MATH_FREE(coefs);
		count = p.count;
		coefs = p.coefs;
		p.count = 0;
		p.coefs = NULL;
		return *this;
	}

	inline double operator() (double x) const {
		double xx = 1.0;
		double cx = 0.0;
		for( size_t i=0; i<count; i++ ) {
			cx += xx * coefs[i];
			xx *= x;
		}
		return cx;
	}

	inline Polynomial operator+ (const Polynomial &p) const {
		Polynomial r;
		r.count = max( count, p.count );
		//r.coefs = new double[r.count];
		r.coefs = MATH_ALLOCS(double,r.count);
		for( size_t i=0; i<r.count; i++ ) {
			r.coefs[i] = ( i<count ? coefs[i] : 0.0 )
				+ ( i<p.count ? p.coefs[i] : 0.0 );
		}
		return r;
	}

	inline Polynomial operator- (const Polynomial &p) const {
		Polynomial r;
		r.count = max( count, p.count );
		//r.coefs = new double[r.count];
		r.coefs = MATH_ALLOCS(double,r.count);
		for( size_t i=0; i<r.count; i++ ) {
			r.coefs[i] = ( i<count ? coefs[i] : 0.0 )
				- ( i<p.count ? p.coefs[i] : 0.0 );
		}
		return r;
	}

	inline Polynomial operator* (const Polynomial &p) const {
		Polynomial r;
		r.count = count * p.count;
		//r.coefs = new double[r.count];
		r.coefs = MATH_ALLOCS(double,r.count);
		for( size_t i=0; i<r.count; i++ ) {
			double cc = 0.0;
			for( size_t j=0; j<=i; j++ ) {
				if( j<count && i-j<p.count )
					cc += coefs[j] * p.coefs[i-j];
			}
			r.coefs[i] = cc;
		}
		return r;
	}

	inline double operator[] (size_t i) const {
		return coefs[i];
	}

	inline size_t power() const {
		return count - 1;
	}

	inline Polynomial integrate() const {
		Polynomial r;
		r.count = count + 1;
		//r.coefs = new double[r.count];
		r.coefs = MATH_ALLOCS(double,r.count);
		r.coefs[0] = 0.0;
		for( size_t i=1; i<r.count; i++ )
			r.coefs[i] = r.coefs[i-1] / i;
		return r;
	}

	inline Polynomial derivate() const {
		Polynomial r;
		r.count = count - 1;
		//r.coefs = new double[r.count];
		r.coefs = MATH_ALLOCS(double,r.count);
		for( size_t i=0; i<r.count; i++ )
			r.coefs[i] = (i+1) * r.coefs[i+1];
		return r;
	}
};

};