#pragma once

namespace csrmath {

template<typename T, size_t W, size_t H> class TBMatrix;

template<typename T, size_t W, size_t H>
class TMatrix {
	T *data;

public:
	inline TMatrix(T* _data) : data(_data) { }
	inline ~TMatrix() { }

	/*inline TMatrix(const TMatrix& mat) : data( new T[W*H] ) {
		for( size_t i=0; i<W*H; i++ )
			data[i] = mat.data[i];
	}*/

	inline size_t width() const { return W; }
	inline size_t height() const { return H; }

	inline T& elem(size_t y, size_t x) {
		return data[ x+W*y ];
	}

	inline const T& elem(size_t y, size_t x) const {
		return data[ x+W*y ];
	}

	inline TBMatrix<T,W,H> operator+( const TMatrix& rarg ) const;
	inline TBMatrix<T,W,H> operator-( const TMatrix& rarg ) const;
	template<size_t W2, size_t H2>
	inline TBMatrix<T,W2,H> operator*( const TMatrix<T,W2,H2>& rarg ) const;

	inline void setIdentity() {
		for( size_t x=0; x<W; x++ )
		for( size_t y=0; y<H; y++ ) {
			data[x+W*y] = ( x==y ? 1 : 0 );
		}
	}

	inline TBMatrix<T,H,W> transpose() const {
		TBMatrix<T,H,W> res;
		for( size_t x=0; x<H; x++ )
		for( size_t y=0; y<W; y++ ) {
			res.elem(y,x) = elem(x,y);
		}
	}

	class TMatrixLine {
		TMatrix &mat;
		size_t Y;
	public:
		TMatrixLine(TMatrix &m, size_t y) : mat(m), Y(y) { }
		inline T& operator[] (size_t x) {
			return mat.elem(Y,x);
		}
	};

	inline TMatrixLine operator[] (size_t y) {
		return TMatrixLine(*this,y);
	}

	friend TBMatrix<T,W,H>;
};

template<typename T, size_t W, size_t H>
class TBMatrix : public TMatrix<T,W,H> {
public:
	//inline TBMatrix() : TMatrix( new T[W*H] ) {
	inline TBMatrix() : TMatrix( MATH_ALLOCS(T,W*H) ) {
		for( size_t i=0; i<W*H; i++ )
			data[i] = 0;
	}
	inline ~TBMatrix() {
		//delete[] data;
		MATH_FREE(data);
	}
	inline TBMatrix(const TMatrix& mat) : TMatrix( MATH_ALLOCS(T,W*H) ) {
		for( size_t i=0; i<W*H; i++ )
			data[i] = mat.data[i];
	}
	inline TBMatrix(const TBMatrix& mat) : TMatrix( MATH_ALLOCS(T,W*H) ) {
		for( size_t i=0; i<W*H; i++ )
			data[i] = mat.data[i];
	}
	inline TBMatrix(TBMatrix&& mat) : TMatrix( mat.data ) {
		mat.data = 0;
	}
};

template<typename T, size_t W, size_t H>
inline TBMatrix<T,W,H> TMatrix<T,W,H>::operator+( const TMatrix& rarg ) const {
	TBMatrix<T,W,H> res;
	for( size_t i=0; i<W*H; i++ )
		res.data[i] = data[i] + rarg.data[i];
	return res;
}

template<typename T, size_t W, size_t H>
inline TBMatrix<T,W,H> TMatrix<T,W,H>::operator-( const TMatrix& rarg ) const {
	TBMatrix<T,W,H> res;
	for( size_t i=0; i<W*H; i++ )
		res.data[i] = data[i] - rarg.data[i];
	return res;
}

template<typename T, size_t W, size_t H>
template<size_t W2, size_t H2>
inline TBMatrix<T,W2,H> TMatrix<T,W,H>::operator*( const TMatrix<T,W2,H2>& rarg ) const {
	if( W!=H2 ) throw 0;
	TBMatrix<T,W2,H> res;
	for( size_t x=0; x<W2; x++ )
	for( size_t y=0; y<H; y++ ) {
		T sum = 0;
		for( size_t i=0; i<W; i++ )
			sum += elem(y,x) * rarg.elem(i,x);
		res.elem(y,x) = sum;
	}
	return res;
}

#ifdef _IOSTREAM_
template<typename T, size_t W, size_t H>
std::ostream& operator<< (std::ostream &sout, const TMatrix<T,W,H>& mat) {
	for( size_t y=0; y<H; y++ ) {
		for( size_t x=0; x<W; x++ )
			sout << mat.elem(y,x) << ( x==W-1 ? '\n' : ' ' );
	}
	return sout;
}
#endif

};
