#include <csrmath.h>
#include <csrmath_vector.h>
#include <csrmath_minimize.h>

using namespace csrmath;

struct csrmath::Minimizer_Ctx {
	double *arg_start;
	double *arg_min;
	double *arg_max;
};

Minimizer::Minimizer(size_t nargs) {
	ctx = MATH_ALLOC(Minimizer_Ctx);
	arg_count = nargs;
	ctx->arg_start = MATH_ALLOCS(double,nargs);
	ctx->arg_min   = MATH_ALLOCS(double,nargs);
	ctx->arg_max   = MATH_ALLOCS(double,nargs);
}

Minimizer::~Minimizer() {
	MATH_FREE(ctx->arg_start);
	MATH_FREE(ctx->arg_min);
	MATH_FREE(ctx->arg_max);
	MATH_FREE(ctx);
}

bool Minimizer::setFunction(MinimizableFunction* _func) {
	func =_func;
	return true;
}

bool Minimizer::setArgument(size_t n, double min, double max, double start) {
	ctx->arg_start[n] = start;
	ctx->arg_min[n]   = min;
	ctx->arg_max[n]   = max;
	return true;
}

VectorD Minimizer::minimize() {
	VectorD arg(arg_count);
	double *arg_step = MATH_ALLOCS(double,arg_count);
	for( size_t i=0; i<arg_count; i++ ) {
		arg[i] = ctx->arg_start[i];
		arg_step[i] = ( ctx->arg_max[i] - ctx->arg_min[i] ) / 20.0;
	}
	size_t cur_arg = 0;
	double prev_val = (*func)(arg);
	for(;;) {
		bool flag0 = true;
		for( size_t i=0; i<arg_count; i++ )
			flag0 &= arg_step[i] < ( ctx->arg_max[i] - ctx->arg_min[i] ) / 1000.0;
		if( flag0 ) break;
		VectorD prev_arg = arg;
		size_t start_arg = cur_arg;
		size_t gstep_cnt = 0;
		do {
			double dir = 1.0;
			size_t dir_cnt = 0;
			while( dir_cnt < 2 ) {
				double val = prev_val;
				size_t step_cnt = 0;
				do {
					prev_val = val;
					prev_arg = arg;
					if( arg[cur_arg] + dir * arg_step[cur_arg] > ctx->arg_max[cur_arg] )
						break;
					if( arg[cur_arg] + dir * arg_step[cur_arg] < ctx->arg_min[cur_arg] )
						break;
					arg[cur_arg] += dir * arg_step[cur_arg];
					val = (*func)(arg);
					step_cnt ++;
				} while( val < prev_val );
				if( step_cnt > 1 )
					gstep_cnt ++;
				arg = prev_arg;
				dir_cnt ++;
				dir = - dir;
			}
			cur_arg = (cur_arg+1) % arg_count;
		} while( cur_arg!=start_arg );
		if( gstep_cnt==0 ) {
			for( size_t i=0; i<arg_count; i++ )
				arg_step[i] /= 10.0;
		}
	}
	MATH_FREE(arg_step);
	return arg;
}
